#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from distutils.core import setup

setup(
    name='justamodel',
    packages=['justamodel'],
    version='0.0.1',
    description='Simple model schema definition',
    url='https://github.com/martin-sucha/justamodel',
    author='Martin Sucha',
    author_email='sucha14@uniba.sk',
    license='MIT',
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: MIT License',
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ]
)